#----------------------------------------------------------------------------#
# Copyright 2010, Regular Polygon

# Permission to use this software without fee is hereby granted.
# Distribution or publication of this software, or substantial portions of
# the software, is subject to the expressed written consent of the author.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#----------------------------------------------------------------------------#
# Name        :   Menger Sponge
# Filename    :   rp_menger_sponge_v10.rb
# Version     :   1.0.1
# SU Version  :   6.0+
# Type        :   Dialog Box
# Description :   Draws a Menger Sponge, from 1-4 levels. 
# Menu Item   :   Draw -> Menger Sponge
# Context Menu:   N/A
# Usage       :   Fill in the dialog box.
# Date        :   8/3/2010
# Author      :   Regular Polygon
# Email       :   regular.polygon@gmail.com
# Website     :   http://regularpolygon.blogspot.com/
#----------------------------------------------------------------------------#


require 'sketchup.rb'
module RP; end


#----------------------------------------------------------------------------#
module RP::MengerSponge

	@@color = ["red", "gold", "LimeGreen", "CornflowerBlue", "BlueViolet", "white"]  

	# Default dialog box parameters
	@@length = (Sketchup.read_default 'RP_menger_sponge', 'length', '81').to_f.inch
	@@level =  (Sketchup.read_default 'RP_menger_sponge', 'level', '2').to_i


	# points for a unit cube centered at origin 
	@@cube_pts = [
		[-0.5,-0.5,-0.5], [+0.5,-0.5,-0.5], [+0.5,+0.5,-0.5], [-0.5,+0.5,-0.5], # bottom
		[-0.5,-0.5,+0.5], [+0.5,-0.5,+0.5], [+0.5,+0.5,+0.5], [-0.5,+0.5,+0.5]  # top
	]

	# face to points association
	@@face_indices = [
		[0, 3, 2, 1],  # -z
		[1, 2, 6, 5],  # +x
		[2, 3, 7, 6],  # +y
		[0, 4, 7, 3],  # -x
		[0, 1, 5, 4],  # -y
		[4, 5, 6, 7]   # +z
	]


	# Component definition for a unit cube
	def self.cube

		# Get the definition if it exists
		cube = Sketchup.active_model.definitions['RP Menger Sponge 0']
		return cube  if cube   

		# Doesn't exist, create definition
		cube = Sketchup.active_model.definitions.add 'RP Menger Sponge 0'
		entities = cube.entities
		@@face_indices.each {|indices|  
			pts = indices.collect {|i| @@cube_pts[i] } 
			entities.add_face pts
		}

		# color 3 sides
		self.color_comp(entities, @@color[0])

		cube
	end


	# 20 transformations that build level n from level n-1 sub-components
	@@trans = []   
	r = Geom::Transformation.new
	scale = Geom::Transformation.new( 1.0 / 3.0 )
	rz = Geom::Transformation.new(ORIGIN, Z_AXIS, 90.degrees)
	ry = Geom::Transformation.new(ORIGIN, Y_AXIS, 90.degrees)
	t1 = Geom::Transformation.new( [-1, -1, -1] )
	t2 = Geom::Transformation.new( [-1, +0, -1] )
	t3 = Geom::Transformation.new( [-1, -1, +0] )
	t4 = Geom::Transformation.new( [-1, -1, +1] ) * ry
	t5 = Geom::Transformation.new( [-1, +0, +1] ) * ry

	for i in 1..4
		@@trans.push scale * r * t1
		@@trans.push scale * r * t2
		@@trans.push scale * r * t3
		@@trans.push scale * r * t4
		@@trans.push scale * r * t5
		r *= rz
	end


	# Component definition for level n of Menger Sponge
	def self.component(n = 0)

		# Base case level 0
		return self.cube  if ( n == 0 )

		# Get the definition if it exists
		d = Sketchup.active_model.definitions["RP Menger Sponge #{ n }"]
		return d  if d   

		# Doesn't exist, create the definition
		d = Sketchup.active_model.definitions.add "RP Menger Sponge #{ n }"
		entities = d.entities

		# build the definition from subcomponents
		d1 = self.component(n-1) 
		components = []
		@@trans.each {|t| components.push( entities.add_instance(d1, t) ) }

		# Now some hacks to reduce geometry.  First explode components.
		# This deletes redundant geometry and gets rid of component profiles.
		components.each {|comp| comp.explode }

		# These edges are always coplanar, and the other face is hidden -- delete all
		edges = entities.find_all {|e|  e.kind_of?(Sketchup::Edge) && e.faces.length == 3 }
		entities.erase_entities edges

		# Now we are going to put geometry back into components grouped by material.
		# Collect all colored faces -- they will all be the same color.
		faces = entities.find_all {|f|  f.kind_of?(Sketchup::Face) && f.material }
		# and copy them to a component.
		d1 = self.copy_faces(faces, "RP Menger Sponge #{ n }-#{ n }")

		# delete the original faces and edges
		entities.erase_entities faces
		edges = entities.find_all {|e|  e.kind_of?(Sketchup::Edge) && e.faces.length == 0 }
		entities.erase_entities edges

		# add a sub-component with same geometry and color
		inst = entities.add_instance(d1, Geom::Transformation.new)
		inst.material = @@color[n-1] 

		# Now group the level n-1 subcomponents leftover from the explosion.  
		m = n - 1
		while (m > 0)
			# collect the instances of the m-th subcomponent
			d1 = Sketchup.active_model.definitions["RP Menger Sponge #{ n-1 }-#{ m }"]
			components = d1.instances.find_all {|inst|  inst.parent == d }

			# Create a new component definition, and copy each instance to it.
			d2 = self.copy_components(components, "RP Menger Sponge #{ n }-#{ m }")

			# delete the original components
			entities.erase_entities components

			# add a sub-component with the copies, and make it the same color
			inst = entities.add_instance(d2, Geom::Transformation.new)
			inst.material = @@color[m-1]  

			m -= 1
		end

		# color 3 sides
		self.color_comp(entities, @@color[n])

		d
	end


	# Create a component definition with a copy of the faces
	def self.copy_faces(faces, def_name)

		# Get the definition if it exists
		d1 = Sketchup.active_model.definitions[def_name]
		return d1  if d1   

		# Doesn't exist, create the definition
		d1 = Sketchup.active_model.definitions.add def_name

		# copy faces
		for f in faces
			pts = f.outer_loop.vertices.collect {|v| v.position }
			d1.entities.add_face pts

			# inner loops are holes in face
			for loop in f.loops                      
				next  if loop == f.outer_loop
				pts = loop.vertices.collect {|v| v.position }
				face = d1.entities.add_face pts
				face.erase!
			end
		end

		d1
	end


	# Create a component definition with a copy of component instances.
	def self.copy_components(components, def_name)

		# Get the definition if it exists
		d1 = Sketchup.active_model.definitions[def_name]
		return d1  if d1   

		# Doesn't exist, create the definition, and copy the instances
		d1 = Sketchup.active_model.definitions.add def_name
		components.each {|inst|  
			d1.entities.add_instance(inst.definition, inst.transformation) 
		}

		d1
	end


	# Color 3 outside faces -- they become the inside color of the next level.
	def self.color_comp(entities, color)
		entities.each {|f|
			next  unless ( f.kind_of?(Sketchup::Face) )
			if ( f.classify_point([0.5, 0.5, 0.5]) == Sketchup::Face::PointOnVertex )
				f.material = color 
			end
		}
	end


	# This method handles the Draw -> Menger Sponge menu item.
	def self.dialog

		# Show dialog to get user input for Menger Sponge paramters
		prompts = ['边长: ', '级别: ']
		defaults = [@@length, @@level]
		list = [nil, '1|2|3|4']
		results = nil

		# Keep prompting user until input is valid.
		begin
			results = UI.inputbox prompts, defaults, list, '门格尔海绵参数'
			return unless results  # user cancelled the operation
			length, level = results
			raise "长度无效: #{ length }"  if ( length <= 0 )
		rescue
			UI.messagebox $!.message
			retry
		end
		
		# Remember parameters
		@@length, @@level = results

		# Remember parameters between sessions
		Sketchup.write_default 'RP_menger_sponge', 'length', @@length.inspect
		Sketchup.write_default 'RP_menger_sponge', 'level', @@level.to_s

		#create the geometry
		self.draw(@@level)
	end


	def self.draw(n = 0)
		return if n > 5

		# make component definition creation undoable
		Sketchup.active_model.start_operation "Menger Sponge"  

			d = Sketchup.active_model.definitions.add 'Menger Sponge'
			entities = d.entities

			d1 = self.component(n)
			scale = Geom::Transformation.new(@@length.to_f)
			inst = entities.add_instance(d1, scale)
			inst.explode

			# remove material from outside faces -- let user paint them how they want.
			entities.each {|f|  
				next  unless f.kind_of?(Sketchup::Face) 
				f.material = f.back_material = nil 
			}

			# Move insertion point to a bottom corner
			x = -@@length.to_f / 2.0
			d.insertion_point = Geom::Point3d.new(x, x, x)

		Sketchup.active_model.commit_operation 

		# place instance in model
		Sketchup.active_model.place_component( d )
	end

end  # module RP::MengerSponge


# Menu
#----------------------------------------------------------------------------#
unless file_loaded? File.basename(__FILE__) 
	# add_separator_to_menu("Draw")
	# UI.menu("Draw").add_item("Menger Sponge") { RP::MengerSponge.dialog } 
	path = Sketchup.find_support_file("Plugins/rp_menger_sponge")
  MengerSponge_tb = UI::Toolbar.new("门格尔海绵参数")
  cmd1 = UI::Command.new("门格尔海绵参数") { RP::MengerSponge.dialog } 
  cmd1.small_icon = File.join(path,"icons/Menger Sponge24.jpg")
  cmd1.large_icon = File.join(path,"icons/Menger Sponge36.jpg")
  cmd1.tooltip = "门格尔海绵参数"
  cmd1.status_bar_text = "门格尔海绵参数"
  MengerSponge_tb.add_item(cmd1)
  MengerSponge_tb.show
end

file_loaded File.basename(__FILE__)